# -*- coding: utf-8 -*-
import base64
import hashlib
import io
import json
import qrcode
import random

from odoo import http, models
from odoo.http import request
from odoo.tools.misc import file_open

CHALLENGE_NS = '/jobs/challenge'


class IrHttp(models.AbstractModel):
    _inherit = 'ir.http'

    @classmethod
    def _auth_method_challenge(cls):
        return request.session.get('challenge_uid')


class Challenge(http.Controller):

    @http.route(CHALLENGE_NS, type='http', auth='public', readonly=True)
    def challenge_index(self, **kw):
        if self._get_user():
            return request.redirect(CHALLENGE_NS + '/next')
        return request.render('odoo_challenge.challenge_login', {}, headers=[('Content-Type', 'text/html; charset=utf-8')])

    @http.route(CHALLENGE_NS + '/login', type='http', auth='public')
    def challenge_login(self, login, **kw):
        if login.strip():
            user = request.env['challenge.user'].sudo().search([('login', '=', login)])

            if not user:
                user = request.env['challenge.user'].sudo().create({
                    'login': login or False,
                    'country_code': request.geoip.country_code,
                })

            request.session.challenge_uid = user.id
            request.session.challenge_current_level = self._get_user().level_id.next_level().id or request.env['challenge.level'].sudo().search([], limit=1)[0].id
        return request.redirect(CHALLENGE_NS + "/next")

    @http.route(CHALLENGE_NS + '/logout', type='http', auth='challenge', readonly=True)
    def challenge_logout(self, **kw):
        request.session.challenge_uid = False
        request.session.challenge_current_level = False
        resp = request.redirect(CHALLENGE_NS)
        resp.delete_cookie('passwd')
        resp.delete_cookie('It-Is-The-Part-2-Of-Url')
        return resp

    def _get_user(self):
        if request.session.get('challenge_uid'):
            return request.env['challenge.user'].sudo().browse(request.session.challenge_uid)
        return False

    def _get_current_level(self):
        return request.session.challenge_uid and request.env['challenge.level'].sudo().browse(request.session.challenge_current_level) or False

    def _success(self):
        level = request.env['challenge.level'].sudo().browse(request.session.challenge_current_level)
        next_level = level.next_level()
        request.session.challenge_current_level = next_level.id

        if level.send_mail:
            user = self._get_user()
            template = request.env.ref('odoo_challenge.challenge_mail_template').sudo()
            template.send_mail(user.id)

        return request.redirect(CHALLENGE_NS + '/next')

    def _log_attempt(self, give, wait, success=None):
        level = self._get_current_level()
        if level.sequence > self._get_user().level: # don't log again if re-do old level
            request.env['challenge.attempt'].sudo().create({
                'user_id': request.session.challenge_uid,
                'level_id': level.id,
                'proposed': give,
                'solution': wait,
                'success': success is None and (give == wait) or success
            })

    @http.route(CHALLENGE_NS + '/next', type='http', auth='challenge', readonly=True)
    def challenge_next(self, **kw):
        if not self._get_user():
            return request.redirect(CHALLENGE_NS)

        level = self._get_current_level()

        if kw.get('l'):
            if str(kw.get('l')).isdigit():
                ask_level = request.env['challenge.level'].sudo().browse(int(kw.get('l')))
                if not level or (ask_level.sequence <= self._get_user().level): # finished all, or level score >>>
                    level = ask_level
            elif kw.get('l') == 'current':
                level = self._get_user().level_id.next_level()
            request.session.challenge_current_level = level.id

        if level:
            level_controller = getattr(self, 'challenge_q_%s' % level.tech_name)
            return level_controller(level, **kw)
        else:
            return self.challenge_finished()

    @http.route(CHALLENGE_NS + '/submit', type='http', auth='challenge', readonly=False)
    def challenge_submit(self, **kw):
        if not request.session.get('challenge_current_level'):
            return request.redirect(CHALLENGE_NS + '/next')
        level = request.env['challenge.level'].sudo().browse(request.session.challenge_current_level)
        # request.session.challenge_current_level = level.id

        level_controller = getattr(self, 'challenge_a_%s' % level.tech_name, self.challenge_a_common)
        return level_controller(level, **kw)

    @http.route(CHALLENGE_NS + '/challenge.css', type='http', auth='challenge', readonly=True)
    def challenge_js(self, **kw):
        cookies = {}
        js = ''
        if not self._get_current_level():
            return request.make_response(' ', headers=[('Content-Type', 'text/css')])
        if self._get_current_level().tech_name == 'console_log':
            js += "console.debug('--> ', atob('%s'), '<--');" % base64.b64encode(self._challenge_pwd().encode('utf-8')).decode('utf-8').replace('\n', '')

        elif self._get_current_level().tech_name == 'network_request':
            js += "// Password is: %s" % self._challenge_pwd()

        elif self._get_current_level().tech_name == 'read_cookies':
            cookies['passwd'] = "%s" % self._challenge_pwd()

        elif self._get_current_level().tech_name == 'json_prime':
            js += "console.debug('--> ', atob('%s'), '<--');" % base64.b64encode(b'You need to call another url to have the next mission ... which one ? Check in another place...').decode('utf-8').replace('\n', '')
            cookies['It-Is-The-Part-2-Of-Url'] = CHALLENGE_NS + "/<part1>/%s/prime.json" % self._get_user().key[-3:]

        return request.make_response(js, headers=[('Content-Type', 'application/javascript; charset=utf-8')], cookies=cookies)

    def _challenge_pwd(self, level=False):
        level = level or self._get_current_level()
        user = self._get_user()

        if level.tech_name == 'header_request':
            rang = range(1, 7, 2)
            pwd = 'Grow your business with Odoo and more than %s apps' % ''.join([str(ord(str(user.key[_]))) for _ in rang])
        elif level.tech_name == 'read_cookies':
            pwd = 'Odoo-%s' % ''.join([str(ord(str(user.key[_]))) for _ in range(26, 40, 2)])
        else:
            pwd = hashlib.sha1((u"%s-%s" % (user.key, level.tech_name)).encode()).hexdigest()

        if level.tech_name == 'simple_algo':
            pwd = pwd[0:5].upper() + pwd[10:15].lower() + pwd[20:25].upper() + pwd[30:35].lower()

        return pwd

    def challenge_a_common(self, level, **kw):
        pwd = self._challenge_pwd(level)
        success = (kw.get('pwd', '') == pwd)
        self._log_attempt(kw.get('pwd', ''), pwd, success=success)
        return success and self._success() or request.redirect(CHALLENGE_NS + '/next?wrong=oups')

    def challenge_q_input_hidden(self, level, **kw):
        pwd = self._challenge_pwd(level)
        return request.render('odoo_challenge.challenge_input_hidden', {'user': self._get_user(), 'level': level, 'pwd': pwd, 'wrong': kw.get('wrong')})

    def challenge_q_console_log(self, level, **kw):
        return request.render('odoo_challenge.challenge_console_log', {'user': self._get_user(), 'level': level, 'wrong': kw.get('wrong')})

    def challenge_q_network_request(self, level, **kw):
        return request.render('odoo_challenge.challenge_network_request', {'user': self._get_user(), 'level': level, 'wrong': kw.get('wrong')})

    def challenge_q_header_request(self, level, **kw):
        pwd = self._challenge_pwd(level)
        return request.render(
            'odoo_challenge.challenge_header_request',
            {'user': self._get_user(), 'level': level, 'wrong': kw.get('wrong')},
            headers=[('It-Is-The-Long-Secret-That-You-Are-Looking-For', pwd)]
        )

    def challenge_q_read_cookies(self, level, **kw):
        # cookies 'passwd' added in other file to hide it
        return request.render('odoo_challenge.challenge_read_cookies', {'user': self._get_user(), 'level': level, 'wrong': kw.get('wrong')})

    def challenge_q_simple_algo(self, level, **kw):
        idx = '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ'
        pwd = self._challenge_pwd()
        tmp = []
        for c in pwd:
            tmp.append(' %s' % idx.index(c))
        return request.render(
            'odoo_challenge.challenge_simple_algo',
            {'user': self._get_user(), 'level': level, 'wrong': kw.get('wrong'), 'pwd_clear': "".join(tmp)}
        )

    def challenge_q_json_prime(self, level, **kw):
        url_part_1 = CHALLENGE_NS + "/%s/<part2>/prime.json" % self._get_user().key[0:3]
        return request.render(
            'odoo_challenge.challenge_json_prime',
            {'user': self._get_user(), 'level': level, 'wrong': kw.get('wrong')},
            headers=[('It-Is-The-Part-1-Of-Url', url_part_1)]
        )

    @http.route(CHALLENGE_NS + '/<string:part1>/<string:part2>/prime.json', type='http', auth='challenge', readonly=True)
    def json_prime(self, **kw):
        if kw.get('part1', '') != self._get_user().key[0:3] or kw.get('part2', '') != self._get_user().key[-3:]:
            return request.redirect(CHALLENGE_NS + "?retry")

        def isprime(n):
            for i in range(2, n):
                if n % i == 0:
                    return False
            return True

        def encode(s, shift=8):
            ret = []
            for i in s:
                # interchar noise
                while random.random() < 0.75:
                    noise = random.randint(0, 2 ** (shift + 8) - 1)
                    if not isprime(noise):
                        ret.append(noise)
                n = ord(i) << shift
                while 1:
                    # low 8 bit noise
                    j = n + random.randint(0, 2 ** shift - 1)
                    if isprime(j):
                        ret.append(j)
                        break
            return ret

        def decode(l, shift=8):
            return "".join(chr(i >> shift) for i in l if isprime(i))

        def get_json(pwd):
            shift = 8
            #index (starting at 100) from 8 in PI
            decimal_8 = [101, 105, 107, 113, 124, 134, 147, 150, 152, 161, 171, 189, 197, 204, 205, 222, 228, 233, 236, 255, 267, 279, 305, 317, 318, 322, 323, 334, 352, 372, 373, 383]
            challenge = {
                "instructions": "Remove non prime numbers from data then right shift [%sth decimal of the constant quantity that determines the area of a circle by multiplying it by the radius squared] bits from the remaining numbers, it will give you the alphanumeric password of the next level" % random.choice(decimal_8),
                "numbers": [],
            }
            challenge['numbers'] = encode(pwd, shift)
            assert pwd == decode(challenge['numbers'], shift)
            return challenge

        pwd = self._challenge_pwd() # to do force level json_prime
        challenge = get_json(pwd)
        return json.dumps(challenge, indent=4, sort_keys=1)

    def challenge_q_img_iframe(self, level, **kw):
        return request.render(
            'odoo_challenge.challenge_img_iframe',
            {'user': self._get_user(), 'level': level, 'wrong': kw.get('wrong')},
        )

    @http.route(CHALLENGE_NS + '/challenge', type='http', auth='challenge')
    def challenge_img_iframe(self, **kw):
        pwd = self._challenge_pwd()
        txt = file_open('odoo_challenge/static/src/img/challenge.txt', 'rb').read().decode('utf-8')
        # P already added because colored as hint
        pwd_msg = "assword: '%s'" % pwd
        for char in pwd_msg:
            txt = txt.replace(' #ffffff;"> </span>', ' #ffffff;">%s</span>' % char, 1)
        return request.make_response(txt, headers=[('Content-Type', 'text/html; charset=utf-8')])

    def challenge_q_base64_qr(self, level, **kw):
        pwd = self._challenge_pwd()
        qr_code = qrcode.make(pwd)
        stream = io.BytesIO()
        qr_code.save(stream)
        b64 = base64.b64encode(stream.getvalue()).decode()

        return request.render(
            'odoo_challenge.challenge_base64_qr',
            {'user': self._get_user(), 'level': level, 'wrong': kw.get('wrong'), 'b64': b64},
        )

    def challenge_q_57_75(self, level, **kw):
        pwd = self._challenge_pwd()
        pwd_int = (int(''.join(c for c in pwd if c.isdigit())) // 100) * 100 + 70  # 70 = 7 * 5 * 2 // Need to end by 70 for the JS check

        resp = request.render(
            'odoo_challenge.challenge_57_75',
            {'user': self._get_user(), 'level': level, 'wrong': kw.get('wrong'), 'ts': pwd_int},
        )
        resp.set_cookie('X-Odoo', "!Odoo - All your applications in one single solution")

        return resp

    def challenge_q_range_algo(self, level, **kw):
        pwd = self._challenge_pwd()
        ranges = [
            '0, 4, 1',
            '4, 20, 2',
            '5, 30, 3',
            '5, 40, 4',
            '6, 40, 6',
            '7, 8, 1',
            '10, 40, 3',
            '10, 40, 5',
            '15, 40, 6',
            '20, 40, 3',
        ]

        output = ""

        for r in ranges:
            output += '\n>>> print(r(pwd, %s))' % r
            rs = list(map(int, r.split(',')))
            output += '\n' + str(''.join([pwd[i] for i in range(rs[0], rs[1], rs[2])]))

        return request.render(
            'odoo_challenge.challenge_range_algo',
            {'user': self._get_user(), 'level': level, 'wrong': kw.get('wrong'), 'output': output},
        )

    @http.route(CHALLENGE_NS + '/70/<string:to_test>', type='http', auth='challenge', readonly=True)
    def check_pwd(self, to_test, **kw):
        def checkPasswordComputed(pwd, to_test):
            stmnt = "!Odoo - All your applications in one single solution"
            ts = str((int(''.join(c for c in pwd if c.isdigit())) // 100) * 100 + 70) # 70 = 7 * 5 * 2 // Need to end by 70 for the JS check
            multi = 1

            for x in ts[0:5]:
                multi *= ord(stmnt[int(x)])
            pwd_computed = str(multi)[1:5] + stmnt[1:5]
            return ((to_test[:-2] == pwd_computed) and (int(to_test[-2:][0]) * int(to_test[-2:][1]) == 35))
        level = request.env['challenge.level'].sudo().search([('tech_name', '=', '57_75')], limit=1)[0]
        pwd = self._challenge_pwd(level)
        if checkPasswordComputed(pwd, to_test):
            return request.make_response(pwd, status=200)
        else:
            return request.make_response('I Am Right You Are Wrong !', status=418)

    @http.route(CHALLENGE_NS + '/soluces', type='http', auth='user', readonly=True)
    def get_pwds(self, **kw):
        if request.env.user.has_group('hr.group_hr_user') and self._get_user():
            pwds = ''
            levels = request.env['challenge.level'].sudo().search([])
            for l in levels:
                pwds += '%s: %s\n' % (l.tech_name, self._challenge_pwd(l))

            return request.make_response(pwds, headers=[('Content-Type', 'text/plain')])
        else:
            raise request.not_found()

    def challenge_finished(self):
        return request.render(
            'odoo_challenge.challenge_finished',
            {'user': self._get_user(), 'finisher': 1},
        )
